"""Util constants/functions for SkyPilot Controllers."""
import copy
import dataclasses
import enum
import getpass
import os
import tempfile
import typing
from typing import Any, Dict, Iterable, List, Optional, Set

import colorama

from sky import check as sky_check
from sky import clouds
from sky import exceptions
from sky import global_user_state
from sky import resources
from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import cloudflare
from sky.clouds import gcp
from sky.data import data_utils
from sky.data import storage as storage_lib
from sky.jobs import constants as managed_job_constants
from sky.jobs import utils as managed_job_utils
from sky.serve import constants as serve_constants
from sky.serve import serve_utils
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import env_options
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky import task as task_lib
    from sky.backends import cloud_vm_ray_backend

logger = sky_logging.init_logger(__name__)

# Message thrown when APIs sky.jobs.launch(), sky.serve.up() received an invalid
# controller resources spec.
CONTROLLER_RESOURCES_NOT_VALID_MESSAGE = (
    '{controller_type} controller resources is not valid, please check '
    '~/.sky/config.yaml file and make sure '
    '{controller_type}.controller.resources is a valid resources spec. '
    'Details:\n  {err}')

# The placeholder for the local skypilot config path in file mounts.
LOCAL_SKYPILOT_CONFIG_PATH_PLACEHOLDER = 'skypilot:local_skypilot_config_path'


@dataclasses.dataclass
class _ControllerSpec:
    """Spec for skypilot controllers."""
    controller_type: str
    name: str
    # Use a list of strings to support fallback to old names. The list is in the
    # fallback order.
    candidate_cluster_names: List[str]
    in_progress_hint: str
    decline_cancel_hint: str
    _decline_down_when_failed_to_fetch_status_hint: str
    decline_down_for_dirty_controller_hint: str
    _check_cluster_name_hint: str
    default_hint_if_non_existent: str
    connection_error_hint: str
    default_resources_config: Dict[str, Any]

    @property
    def cluster_name(self) -> str:
        """The name in candidate_cluster_names that exists, else the first."""
        for candidate_name in self.candidate_cluster_names:
            record = global_user_state.get_cluster_from_name(candidate_name)
            if record is not None:
                return candidate_name
        return self.candidate_cluster_names[0]

    @property
    def decline_down_when_failed_to_fetch_status_hint(self) -> str:
        return self._decline_down_when_failed_to_fetch_status_hint.format(
            cluster_name=self.cluster_name)

    @property
    def check_cluster_name_hint(self) -> str:
        return self._check_cluster_name_hint.format(
            cluster_name=self.cluster_name)


class Controllers(enum.Enum):
    """Skypilot controllers."""
    # NOTE(dev): Keep this align with
    # sky/cli.py::_CONTROLLER_TO_HINT_OR_RAISE
    JOBS_CONTROLLER = _ControllerSpec(
        controller_type='jobs',
        name='managed jobs controller',
        candidate_cluster_names=[
            managed_job_utils.JOB_CONTROLLER_NAME,
            managed_job_utils.LEGACY_JOB_CONTROLLER_NAME
        ],
        in_progress_hint=(
            '* {job_info}To see all managed jobs: '
            f'{colorama.Style.BRIGHT}sky jobs queue{colorama.Style.RESET_ALL}'),
        decline_cancel_hint=(
            'Cancelling the jobs controller\'s jobs is not allowed.\nTo cancel '
            f'managed jobs, use: {colorama.Style.BRIGHT}sky jobs cancel '
            f'<managed job IDs> [--all]{colorama.Style.RESET_ALL}'),
        _decline_down_when_failed_to_fetch_status_hint=(
            f'{colorama.Fore.RED}Tearing down the jobs controller while '
            'it is in INIT state is not supported (this means a job launch '
            'is in progress or the previous launch failed), as we cannot '
            'guarantee that all the managed jobs are finished. Please wait '
            'until the jobs controller is UP or fix it with '
            f'{colorama.Style.BRIGHT}sky start '
            '{cluster_name}'
            f'{colorama.Style.RESET_ALL}.'),
        decline_down_for_dirty_controller_hint=(
            f'{colorama.Fore.RED}In-progress managed jobs found. To avoid '
            f'resource leakage, cancel all jobs first: {colorama.Style.BRIGHT}'
            f'sky jobs cancel -a{colorama.Style.RESET_ALL}\n'),
        _check_cluster_name_hint=('Cluster {cluster_name} is reserved for '
                                  'managed jobs controller.'),
        default_hint_if_non_existent='No in-progress managed jobs.',
        connection_error_hint=(
            'Failed to connect to jobs controller, please try again later.'),
        default_resources_config=managed_job_constants.CONTROLLER_RESOURCES)
    SKY_SERVE_CONTROLLER = _ControllerSpec(
        controller_type='serve',
        name='serve controller',
        candidate_cluster_names=[serve_utils.SKY_SERVE_CONTROLLER_NAME],
        in_progress_hint=(
            f'* To see detailed service status: {colorama.Style.BRIGHT}'
            f'sky serve status -a{colorama.Style.RESET_ALL}'),
        decline_cancel_hint=(
            'Cancelling the sky serve controller\'s jobs is not allowed.'),
        _decline_down_when_failed_to_fetch_status_hint=(
            f'{colorama.Fore.RED}Tearing down the sky serve controller '
            'while it is in INIT state is not supported (this means a sky '
            'serve up is in progress or the previous launch failed), as we '
            'cannot guarantee that all the services are terminated. Please '
            'wait until the sky serve controller is UP or fix it with '
            f'{colorama.Style.BRIGHT}sky start '
            '{cluster_name}'
            f'{colorama.Style.RESET_ALL}.'),
        decline_down_for_dirty_controller_hint=(
            f'{colorama.Fore.RED}Tearing down the sky serve controller is not '
            'supported, as it is currently serving the following services: '
            '{service_names}. Please terminate the services first with '
            f'{colorama.Style.BRIGHT}sky serve down -a'
            f'{colorama.Style.RESET_ALL}.'),
        _check_cluster_name_hint=('Cluster {cluster_name} is reserved for '
                                  'sky serve controller.'),
        default_hint_if_non_existent='No live services.',
        connection_error_hint=(
            'Failed to connect to serve controller, please try again later.'),
        default_resources_config=serve_constants.CONTROLLER_RESOURCES)

    @classmethod
    def from_name(cls, name: Optional[str]) -> Optional['Controllers']:
        """Check if the cluster name is a controller name.

        Returns:
            The controller if the cluster name is a controller name.
            Otherwise, returns None.
        """
        for controller in cls:
            if name in controller.value.candidate_cluster_names:
                return controller
        return None

    @classmethod
    def from_type(cls, controller_type: str) -> Optional['Controllers']:
        """Get the controller by controller type.

        Returns:
            The controller if the controller type is valid.
            Otherwise, returns None.
        """
        for controller in cls:
            if controller.value.controller_type == controller_type:
                return controller
        return None


# Install cli dependencies. Not using SkyPilot wheels because the wheel
# can be cleaned up by another process.
# TODO(zhwu): Keep the dependencies align with the ones in setup.py
def _get_cloud_dependencies_installation_commands(
        controller: Controllers) -> List[str]:
    # TODO(tian): Make dependency installation command a method of cloud
    # class and get all installation command for enabled clouds.
    commands = []
    prefix_str = 'Check & install cloud dependencies on controller: '
    # This is to make sure the shorter checking message does not have junk
    # characters from the previous message.
    empty_str = ' ' * 10
    aws_dependencies_installation = (
        'pip list | grep boto3 > /dev/null 2>&1 || pip install '
        'botocore>=1.29.10 boto3>=1.26.1; '
        # Need to separate the installation of awscli from above because some
        # other clouds will install boto3 but not awscli.
        'pip list | grep awscli> /dev/null 2>&1 || pip install "urllib3<2" '
        'awscli>=1.27.10 "colorama<0.4.5" > /dev/null 2>&1')
    for cloud in sky_check.get_cached_enabled_clouds_or_refresh():
        if isinstance(
                clouds,
            (clouds.Lambda, clouds.SCP, clouds.Fluidstack, clouds.Paperspace)):
            # no need to install any cloud dependencies for lambda, scp,
            # fluidstack and paperspace
            continue
        if isinstance(cloud, clouds.AWS):
            commands.append(f'echo -en "\\r{prefix_str}AWS{empty_str}" && ' +
                            aws_dependencies_installation)
        elif isinstance(cloud, clouds.Azure):
            commands.append(
                f'echo -en "\\r{prefix_str}Azure{empty_str}" && '
                'pip list | grep azure-cli > /dev/null 2>&1 || '
                'pip install "azure-cli>=2.31.0" azure-core '
                '"azure-identity>=1.13.0" azure-mgmt-network > /dev/null 2>&1')
            # Have to separate this installation of az blob storage from above
            # because this is newly-introduced and not part of azure-cli. We
            # need a separate installed check for this.
            commands.append(
                'pip list | grep azure-storage-blob > /dev/null 2>&1 || '
                'pip install azure-storage-blob msgraph-sdk > /dev/null 2>&1')
        elif isinstance(cloud, clouds.GCP):
            commands.append(
                f'echo -en "\\r{prefix_str}GCP{empty_str}" && '
                'pip list | grep google-api-python-client > /dev/null 2>&1 || '
                'pip install "google-api-python-client>=2.69.0" '
                '> /dev/null 2>&1')
            # Have to separate the installation of google-cloud-storage from
            # above because for a VM launched on GCP, the VM may have
            # google-api-python-client installed alone.
            commands.append(
                'pip list | grep google-cloud-storage > /dev/null 2>&1 || '
                'pip install google-cloud-storage > /dev/null 2>&1')
            commands.append(f'{gcp.GOOGLE_SDK_INSTALLATION_COMMAND}')
        elif isinstance(cloud, clouds.Kubernetes):
            commands.append(
                f'echo -en "\\r{prefix_str}Kubernetes{empty_str}" && '
                'pip list | grep kubernetes > /dev/null 2>&1 || '
                'pip install "kubernetes>=20.0.0" > /dev/null 2>&1 &&'
                # Install k8s + skypilot dependencies
                'sudo bash -c "if '
                '! command -v curl &> /dev/null || '
                '! command -v socat &> /dev/null || '
                '! command -v netcat &> /dev/null; '
                'then apt update && apt install curl socat netcat -y '
                '&> /dev/null; '
                'fi" && '
                # Install kubectl
                '(command -v kubectl &>/dev/null || '
                '(curl -s -LO "https://dl.k8s.io/release/'
                '$(curl -L -s https://dl.k8s.io/release/stable.txt)'
                '/bin/linux/amd64/kubectl" && '
                'sudo install -o root -g root -m 0755 '
                'kubectl /usr/local/bin/kubectl))')
        elif isinstance(cloud, clouds.Cudo):
            commands.append(
                f'echo -en "\\r{prefix_str}Cudo{empty_str}" && '
                'pip list | grep cudo-compute > /dev/null 2>&1 || '
                'pip install "cudo-compute>=0.1.10" > /dev/null 2>&1 && '
                'wget https://download.cudo.org/compute/cudoctl-0.3.2-amd64.deb -O ~/cudoctl.deb > /dev/null 2>&1 && '  # pylint: disable=line-too-long
                'sudo dpkg -i ~/cudoctl.deb > /dev/null 2>&1')
        if controller == Controllers.JOBS_CONTROLLER:
            if isinstance(cloud, clouds.IBM):
                commands.append(
                    f'echo -en "\\r{prefix_str}IBM{empty_str}" '
                    '&& pip list | grep ibm-cloud-sdk-core > /dev/null 2>&1 || '
                    'pip install ibm-cloud-sdk-core ibm-vpc '
                    'ibm-platform-services ibm-cos-sdk > /dev/null 2>&1')
            elif isinstance(cloud, clouds.OCI):
                commands.append(f'echo -en "\\r{prefix_str}OCI{empty_str}" && '
                                'pip list | grep oci > /dev/null 2>&1 || '
                                'pip install oci > /dev/null 2>&1')
            elif isinstance(cloud, clouds.RunPod):
                commands.append(
                    f'echo -en "\\r{prefix_str}RunPod{empty_str}" && '
                    'pip list | grep runpod > /dev/null 2>&1 || '
                    'pip install "runpod>=1.5.1" > /dev/null 2>&1')
    if (cloudflare.NAME
            in storage_lib.get_cached_enabled_storage_clouds_or_refresh()):
        commands.append(f'echo -en "\\r{prefix_str}Cloudflare{empty_str}" && ' +
                        aws_dependencies_installation)
    commands.append(f'echo -e "\\r{prefix_str}Done for {len(commands)} '
                    'clouds."')
    return commands


def check_cluster_name_not_controller(
        cluster_name: Optional[str],
        operation_str: Optional[str] = None) -> None:
    """Errors out if the cluster name is a controller name.

    Raises:
      sky.exceptions.NotSupportedError: if the cluster name is a controller
        name, raise with an error message explaining 'operation_str' is not
        allowed.

    Returns:
      None, if the cluster name is not a controller name.
    """
    controller = Controllers.from_name(cluster_name)
    if controller is not None:
        msg = controller.value.check_cluster_name_hint
        if operation_str is not None:
            msg += f' {operation_str} is not allowed.'
        with ux_utils.print_exception_no_traceback():
            raise exceptions.NotSupportedError(msg)


# Internal only:
def download_and_stream_latest_job_log(
        backend: 'cloud_vm_ray_backend.CloudVmRayBackend',
        handle: 'cloud_vm_ray_backend.CloudVmRayResourceHandle',
        local_dir: str) -> Optional[str]:
    """Downloads and streams the latest job log.

    This function is only used by jobs controller and sky serve controller.
    """
    os.makedirs(local_dir, exist_ok=True)
    log_file = None
    try:
        log_dirs = backend.sync_down_logs(
            handle,
            # Download the log of the latest job.
            # The job_id for the managed job running on the cluster is not
            # necessarily 1, as it is possible that the worker node in a
            # multi-node cluster is preempted, and we recover the managed job
            # on the existing cluster, which leads to a larger job_id. Those
            # job_ids all represent the same logical managed job.
            job_ids=None,
            local_dir=local_dir)
    except exceptions.CommandError as e:
        logger.info(f'Failed to download the logs: '
                    f'{common_utils.format_exception(e)}')
    else:
        if not log_dirs:
            logger.error('Failed to find the logs for the user program.')
        else:
            log_dir = list(log_dirs.values())[0]
            log_file = os.path.join(log_dir, 'run.log')

            # Print the logs to the console.
            try:
                with open(log_file, 'r', encoding='utf-8') as f:
                    print(f.read())
            except FileNotFoundError:
                logger.error('Failed to find the logs for the user '
                             f'program at {log_file}.')
    return log_file


def shared_controller_vars_to_fill(
        controller: Controllers,
        remote_user_config_path: str) -> Dict[str, str]:
    vars_to_fill: Dict[str, Any] = {
        'cloud_dependencies_installation_commands':
            _get_cloud_dependencies_installation_commands(controller),
        # We need to activate the python environment on the controller to ensure
        # cloud SDKs are installed in SkyPilot runtime environment and can be
        # accessed.
        'sky_activate_python_env': constants.ACTIVATE_SKY_REMOTE_PYTHON_ENV,
        'sky_python_cmd': constants.SKY_PYTHON_CMD,
    }
    env_vars: Dict[str, str] = {
        env.value: '1' for env in env_options.Options if env.get()
    }
    env_vars.update({
        # Should not use $USER here, as that env var can be empty when
        # running in a container.
        constants.USER_ENV_VAR: getpass.getuser(),
        constants.USER_ID_ENV_VAR: common_utils.get_user_hash(),
        # Skip cloud identity check to avoid the overhead.
        env_options.Options.SKIP_CLOUD_IDENTITY_CHECK.value: '1',
    })
    if skypilot_config.loaded():
        # Only set the SKYPILOT_CONFIG env var if the user has a config file.
        env_vars[
            skypilot_config.ENV_VAR_SKYPILOT_CONFIG] = remote_user_config_path
    vars_to_fill['controller_envs'] = env_vars
    return vars_to_fill


def get_controller_resources(
    controller: Controllers,
    task_resources: Iterable['resources.Resources'],
) -> Set['resources.Resources']:
    """Read the skypilot config and setup the controller resources.

    Returns:
        A set of controller resources that will be used to launch the
        controller. All fields are the same except for the cloud. If no
        controller exists and the controller resources has no cloud
        specified, the controller will be launched on one of the clouds
        of the task resources for better connectivity.
    """
    controller_resources_config_copied: Dict[str, Any] = copy.copy(
        controller.value.default_resources_config)
    if skypilot_config.loaded():
        # Override the controller resources with the ones specified in the
        # config.
        custom_controller_resources_config = skypilot_config.get_nested(
            (controller.value.controller_type, 'controller', 'resources'), None)
        if custom_controller_resources_config is not None:
            controller_resources_config_copied.update(
                custom_controller_resources_config)
        elif controller == Controllers.JOBS_CONTROLLER:
            controller_resources_config_copied.update(
                skypilot_config.get_nested(('spot', 'controller', 'resources'),
                                           {}))

    try:
        controller_resources = resources.Resources.from_yaml_config(
            controller_resources_config_copied)
    except ValueError as e:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                CONTROLLER_RESOURCES_NOT_VALID_MESSAGE.format(
                    controller_type=controller.value.controller_type,
                    err=common_utils.format_exception(
                        e, use_bracket=True)).capitalize()) from e
    # TODO(tian): Support multiple resources for the controller. One blocker
    # here is the semantic if controller resources use `ordered` and we want
    # to override it with multiple cloud from task resources.
    if len(controller_resources) != 1:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                CONTROLLER_RESOURCES_NOT_VALID_MESSAGE.format(
                    controller_type=controller.value.controller_type,
                    err=f'Expected exactly one resource, got '
                    f'{len(controller_resources)} resources: '
                    f'{controller_resources}').capitalize())
    controller_resources_to_use: resources.Resources = list(
        controller_resources)[0]

    controller_record = global_user_state.get_cluster_from_name(
        controller.value.cluster_name)
    if controller_record is not None:
        handle = controller_record.get('handle', None)
        if handle is not None:
            controller_resources_to_use = handle.launched_resources

    if controller_resources_to_use.cloud is not None:
        return {controller_resources_to_use}

    # If the controller and replicas are from the same cloud, it should
    # provide better connectivity. We will let the controller choose from
    # the clouds of the resources if the controller does not exist.
    # TODO(tian): Consider respecting the regions/zones specified for the
    # resources as well.
    requested_clouds: Set['clouds.Cloud'] = set()
    for resource in task_resources:
        # cloud is an object and will not be able to be distinguished by set.
        # Here we manually check if the cloud is in the set.
        if resource.cloud is not None:
            if not clouds.cloud_in_iterable(resource.cloud, requested_clouds):
                try:
                    resource.cloud.check_features_are_supported(
                        resources.Resources(),
                        {clouds.CloudImplementationFeatures.HOST_CONTROLLERS})
                except exceptions.NotSupportedError:
                    # Skip the cloud if it does not support hosting controllers.
                    continue
                requested_clouds.add(resource.cloud)
        else:
            # if one of the resource.cloud is None, this could represent user
            # does not know which cloud is best for the specified resources.
            # For example:
            #   resources:
            #     - accelerators: L4     # Both available on AWS and GCP
            #     - cloud: runpod
            #       accelerators: A40
            # In this case, we allow the controller to be launched on any cloud.
            requested_clouds.clear()
            break
    if not requested_clouds:
        return {controller_resources_to_use}
    return {
        controller_resources_to_use.copy(cloud=controller_cloud)
        for controller_cloud in requested_clouds
    }


def _setup_proxy_command_on_controller(
        controller_launched_cloud: 'clouds.Cloud') -> Dict[str, Any]:
    """Sets up proxy command on the controller.

    This function should be called on the controller (remote cluster), which
    has the `~/.sky/sky_ray.yaml` file.
    """
    # Look up the contents of the already loaded configs via the
    # 'skypilot_config' module. Don't simply read the on-disk file as
    # it may have changed since this process started.
    #
    # Set any proxy command to None, because the controller would've
    # been launched behind the proxy, and in general any nodes we
    # launch may not have or need the proxy setup. (If the controller
    # needs to launch mew clusters in another region/VPC, the user
    # should properly set up VPC peering, which will allow the
    # cross-region/VPC communication. The proxy command is orthogonal
    # to this scenario.)
    #
    # This file will be uploaded to the controller node and will be
    # used throughout the managed job's / service's recovery attempts
    # (i.e., if it relaunches due to preemption, we make sure the
    # same config is used).
    #
    # NOTE: suppose that we have a controller in old VPC, then user
    # changes 'vpc_name' in the config and does a 'job launch' /
    # 'serve up'. In general, the old controller may not successfully
    # launch the job in the new VPC. This happens if the two VPCs don’t
    # have peering set up. Like other places in the code, we assume
    # properly setting up networking is user's responsibilities.
    # TODO(zongheng): consider adding a basic check that checks
    # controller VPC (or name) == the managed job's / service's VPC
    # (or name). It may not be a sufficient check (as it's always
    # possible that peering is not set up), but it may catch some
    # obvious errors.
    proxy_command_key = (str(controller_launched_cloud).lower(),
                         'ssh_proxy_command')
    ssh_proxy_command = skypilot_config.get_nested(proxy_command_key, None)
    config_dict = skypilot_config.to_dict()
    if isinstance(ssh_proxy_command, str):
        config_dict = skypilot_config.set_nested(proxy_command_key, None)
    elif isinstance(ssh_proxy_command, dict):
        # Instead of removing the key, we set the value to empty string
        # so that the controller will only try the regions specified by
        # the keys.
        ssh_proxy_command = {k: None for k in ssh_proxy_command}
        config_dict = skypilot_config.set_nested(proxy_command_key,
                                                 ssh_proxy_command)

    return config_dict


def replace_skypilot_config_path_in_file_mounts(
        cloud: 'clouds.Cloud', file_mounts: Optional[Dict[str, str]]):
    """Replaces the SkyPilot config path in file mounts with the real path."""
    # TODO(zhwu): This function can be moved to `backend_utils` once we have
    # more predefined file mounts that needs to be replaced after the cluster
    # is provisioned, e.g., we may need to decide which cloud to create a bucket
    # to be mounted to the cluster based on the cloud the cluster is actually
    # launched on (after failover).
    if file_mounts is None:
        return
    replaced = False
    to_replace = True
    with tempfile.NamedTemporaryFile('w', delete=False) as f:
        if skypilot_config.loaded():
            new_skypilot_config = _setup_proxy_command_on_controller(cloud)
            common_utils.dump_yaml(f.name, new_skypilot_config)
            to_replace = True
        else:
            # Empty config. Remove the placeholder below.
            to_replace = False
        for remote_path, local_path in list(file_mounts.items()):
            if local_path == LOCAL_SKYPILOT_CONFIG_PATH_PLACEHOLDER:
                if to_replace:
                    file_mounts[remote_path] = f.name
                    replaced = True
                else:
                    del file_mounts[remote_path]
    if replaced:
        logger.debug(f'Replaced {LOCAL_SKYPILOT_CONFIG_PATH_PLACEHOLDER} with '
                     f'the real path in file mounts: {file_mounts}')


def maybe_translate_local_file_mounts_and_sync_up(task: 'task_lib.Task',
                                                  path: str) -> None:
    """Translates local->VM mounts into Storage->VM, then syncs up any Storage.

    Eagerly syncing up local->Storage ensures Storage->VM would work at task
    launch time.

    If there are no local source paths to be translated, this function would
    still sync up any storage mounts with local source paths (which do not
    undergo translation).
    """
    # ================================================================
    # Translate the workdir and local file mounts to cloud file mounts.
    # ================================================================
    run_id = common_utils.get_usage_run_id()[:8]
    original_file_mounts = task.file_mounts if task.file_mounts else {}
    original_storage_mounts = task.storage_mounts if task.storage_mounts else {}

    copy_mounts = task.get_local_to_remote_file_mounts()
    if copy_mounts is None:
        copy_mounts = {}

    has_local_source_paths_file_mounts = bool(copy_mounts)
    has_local_source_paths_workdir = task.workdir is not None

    msg = None
    if has_local_source_paths_workdir and has_local_source_paths_file_mounts:
        msg = 'workdir and file_mounts with local source paths'
    elif has_local_source_paths_file_mounts:
        msg = 'file_mounts with local source paths'
    elif has_local_source_paths_workdir:
        msg = 'workdir'
    if msg:
        logger.info(f'{colorama.Fore.YELLOW}Translating {msg} to SkyPilot '
                    f'Storage...{colorama.Style.RESET_ALL}')

    # Step 1: Translate the workdir to SkyPilot storage.
    new_storage_mounts = {}
    if task.workdir is not None:
        bucket_name = constants.WORKDIR_BUCKET_NAME.format(
            username=common_utils.get_cleaned_username(), id=run_id)
        workdir = task.workdir
        task.workdir = None
        if (constants.SKY_REMOTE_WORKDIR in original_file_mounts or
                constants.SKY_REMOTE_WORKDIR in original_storage_mounts):
            raise ValueError(
                f'Cannot mount {constants.SKY_REMOTE_WORKDIR} as both the '
                'workdir and file_mounts contains it as the target.')
        new_storage_mounts[
            constants.
            SKY_REMOTE_WORKDIR] = storage_lib.Storage.from_yaml_config({
                'name': bucket_name,
                'source': workdir,
                'persistent': False,
                'mode': 'COPY',
            })
        # Check of the existence of the workdir in file_mounts is done in
        # the task construction.
        logger.info(f'Workdir {workdir!r} will be synced to cloud storage '
                    f'{bucket_name!r}.')

    # Step 2: Translate the local file mounts with folder in src to SkyPilot
    # storage.
    # TODO(zhwu): Optimize this by:
    # 1. Use the same bucket for all the mounts.
    # 2. When the src is the same, use the same bucket.
    copy_mounts_with_file_in_src = {}
    for i, (dst, src) in enumerate(copy_mounts.items()):
        assert task.file_mounts is not None
        task.file_mounts.pop(dst)
        if os.path.isfile(os.path.abspath(os.path.expanduser(src))):
            copy_mounts_with_file_in_src[dst] = src
            continue
        bucket_name = constants.FILE_MOUNTS_BUCKET_NAME.format(
            username=common_utils.get_cleaned_username(),
            id=f'{run_id}-{i}',
        )
        new_storage_mounts[dst] = storage_lib.Storage.from_yaml_config({
            'name': bucket_name,
            'source': src,
            'persistent': False,
            'mode': 'COPY',
        })
        logger.info(
            f'Folder in local file mount {src!r} will be synced to SkyPilot '
            f'storage {bucket_name}.')

    # Step 3: Translate local file mounts with file in src to SkyPilot storage.
    # Hard link the files in src to a temporary directory, and upload folder.
    local_fm_path = os.path.join(
        tempfile.gettempdir(),
        constants.FILE_MOUNTS_LOCAL_TMP_DIR.format(id=run_id))
    os.makedirs(local_fm_path, exist_ok=True)
    file_bucket_name = constants.FILE_MOUNTS_FILE_ONLY_BUCKET_NAME.format(
        username=common_utils.get_cleaned_username(), id=run_id)
    file_mount_remote_tmp_dir = constants.FILE_MOUNTS_REMOTE_TMP_DIR.format(
        path)
    if copy_mounts_with_file_in_src:
        src_to_file_id = {}
        for i, src in enumerate(set(copy_mounts_with_file_in_src.values())):
            src_to_file_id[src] = i
            os.link(os.path.abspath(os.path.expanduser(src)),
                    os.path.join(local_fm_path, f'file-{i}'))

        new_storage_mounts[
            file_mount_remote_tmp_dir] = storage_lib.Storage.from_yaml_config({
                'name': file_bucket_name,
                'source': local_fm_path,
                'persistent': False,
                'mode': 'MOUNT',
            })
        if file_mount_remote_tmp_dir in original_storage_mounts:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Failed to translate file mounts, due to the default '
                    f'destination {file_mount_remote_tmp_dir} '
                    'being taken.')
        sources = list(src_to_file_id.keys())
        sources_str = '\n\t'.join(sources)
        logger.info('Source files in file_mounts will be synced to '
                    f'cloud storage {file_bucket_name}:'
                    f'\n\t{sources_str}')
    task.update_storage_mounts(new_storage_mounts)

    # Step 4: Upload storage from sources
    # Upload the local source to a bucket. The task will not be executed
    # locally, so we need to upload the files/folders to the bucket manually
    # here before sending the task to the remote jobs controller.
    if task.storage_mounts:
        # There may be existing (non-translated) storage mounts, so log this
        # whenever task.storage_mounts is non-empty.
        logger.info(f'{colorama.Fore.YELLOW}Uploading sources to cloud storage.'
                    f'{colorama.Style.RESET_ALL} See: sky storage ls')
    try:
        task.sync_storage_mounts()
    except ValueError as e:
        if 'No enabled cloud for storage' in str(e):
            data_src = None
            if has_local_source_paths_file_mounts:
                data_src = 'file_mounts'
            if has_local_source_paths_workdir:
                if data_src:
                    data_src += ' and workdir'
                else:
                    data_src = 'workdir'
            store_enabled_clouds = ', '.join(storage_lib.STORE_ENABLED_CLOUDS)
            with ux_utils.print_exception_no_traceback():
                raise exceptions.NotSupportedError(
                    f'Unable to use {data_src} - no cloud with object store '
                    'is enabled. Please enable at least one cloud with '
                    f'object store support ({store_enabled_clouds}) by running '
                    f'`sky check`, or remove {data_src} from your task.'
                    '\nHint: If you do not have any cloud access, you may still'
                    ' download data and code over the network using curl or '
                    'other tools in the `setup` section of the task.') from None

    # Step 5: Add the file download into the file mounts, such as
    #  /original-dst: s3://spot-fm-file-only-bucket-name/file-0
    new_file_mounts = {}
    if copy_mounts_with_file_in_src:
        # file_mount_remote_tmp_dir will only exist when there are files in
        # the src for copy mounts.
        storage_obj = task.storage_mounts[file_mount_remote_tmp_dir]
        store_type = list(storage_obj.stores.keys())[0]
        store_object = storage_obj.stores[store_type]
        bucket_url = storage_lib.StoreType.get_endpoint_url(
            store_object, file_bucket_name)
        for dst, src in copy_mounts_with_file_in_src.items():
            file_id = src_to_file_id[src]
            new_file_mounts[dst] = bucket_url + f'/file-{file_id}'
    task.update_file_mounts(new_file_mounts)

    # Step 6: Replace the source field that is local path in all storage_mounts
    # with bucket URI and remove the name field.
    for storage_obj in task.storage_mounts.values():
        if (storage_obj.source is not None and
                not data_utils.is_cloud_store_url(storage_obj.source)):
            # Need to replace the local path with bucket URI, and remove the
            # name field, so that the storage mount can work on the spot
            # controller.
            store_types = list(storage_obj.stores.keys())
            assert len(store_types) == 1, (
                'We only support one store type for now.', storage_obj.stores)
            store_type = store_types[0]
            store_object = storage_obj.stores[store_type]
            storage_obj.source = storage_lib.StoreType.get_endpoint_url(
                store_object, storage_obj.name)
            storage_obj.force_delete = True

    # Step 7: Convert all `MOUNT` mode storages which don't specify a source
    # to specifying a source. If the source is specified with a local path,
    # it was handled in step 6.
    updated_mount_storages = {}
    for storage_path, storage_obj in task.storage_mounts.items():
        if (storage_obj.mode == storage_lib.StorageMode.MOUNT and
                not storage_obj.source):
            # Construct source URL with first store type and storage name
            # E.g., s3://my-storage-name
            store_types = list(storage_obj.stores.keys())
            assert len(store_types) == 1, (
                'We only support one store type for now.', storage_obj.stores)
            store_type = store_types[0]
            store_object = storage_obj.stores[store_type]
            source = storage_lib.StoreType.get_endpoint_url(
                store_object, storage_obj.name)
            new_storage = storage_lib.Storage.from_yaml_config({
                'source': source,
                'persistent': storage_obj.persistent,
                'mode': storage_lib.StorageMode.MOUNT.value,
                # We enable force delete to allow the controller to delete
                # the object store in case persistent is set to False.
                '_force_delete': True
            })
            updated_mount_storages[storage_path] = new_storage
    task.update_storage_mounts(updated_mount_storages)



"""FluidStack API client."""

import functools
import json
import os
from typing import Any, Dict, List, Optional
import uuid

import requests


def get_key_suffix():
    return str(uuid.uuid4()).replace('-', '')[:8]


ENDPOINT = 'https://api.fluidstack.io/v1/'
FLUIDSTACK_API_KEY_PATH = '~/.fluidstack/api_key'
FLUIDSTACK_API_TOKEN_PATH = '~/.fluidstack/api_token'


def read_contents(path: str) -> str:
    with open(path, mode='r', encoding='utf-8') as f:
        return f.read().strip()


class FluidstackAPIError(Exception):

    def __init__(self, message: str, code: int = 400):
        self.code = code
        super().__init__(message)


def raise_fluidstack_error(response: requests.Response) -> None:
    """Raise FluidstackAPIError if appropriate."""
    status_code = response.status_code
    if response.ok:
        return
    try:
        resp_json = response.json()
        message = resp_json.get('error', response.text)
    except (KeyError, json.decoder.JSONDecodeError) as e:
        raise FluidstackAPIError(
            f'Unexpected error. Status code: {status_code} \n {response.text}'
            f'\n {str(e)}',
            code=status_code) from e
    raise FluidstackAPIError(f'{message}', status_code)


@functools.lru_cache()
def with_nvidia_drivers(region: str):
    if region in ['norway_4_eu', 'generic_1_canada']:
        return False
    client = FluidstackClient()
    plans = client.get_plans()
    for plan in plans:
        if region in [r['id'] for r in plan['regions']]:
            if 'Ubuntu 20.04 LTS (Nvidia)' in plan['os_options']:
                return True
    return False


class FluidstackClient:
    """FluidStack API Client"""

    def __init__(self):
        self.api_key = read_contents(
            os.path.expanduser(FLUIDSTACK_API_KEY_PATH))
        self.api_token = read_contents(
            os.path.expanduser(FLUIDSTACK_API_TOKEN_PATH))

    def get_plans(self):
        response = requests.get(ENDPOINT + 'plans')
        raise_fluidstack_error(response)
        plans = response.json()
        plans = [
            plan for plan in plans
            if plan['minimum_commitment'] == 'hourly' and plan['type'] in
            ['preconfigured', 'custom'] and plan['gpu_type'] != 'NO GPU'
        ]
        return plans

    def list_instances(
            self,
            tag_filters: Optional[Dict[str,
                                       str]] = None) -> List[Dict[str, Any]]:
        response = requests.get(
            ENDPOINT + 'servers',
            auth=(self.api_key, self.api_token),
        )
        raise_fluidstack_error(response)
        instances = response.json()
        filtered_instances = []

        for instance in instances:
            if isinstance(instance['tags'], str):
                instance['tags'] = json.loads(instance['tags'])
            if not instance['tags']:
                instance['tags'] = {}
            if tag_filters:
                for key in tag_filters:
                    if instance['tags'].get(key, None) != tag_filters[key]:
                        break
                else:
                    filtered_instances.append(instance)
            else:
                filtered_instances.append(instance)

        return filtered_instances

    def create_instance(
        self,
        instance_type: str = '',
        hostname: str = '',
        region: str = '',
        ssh_pub_key: str = '',
        count: int = 1,
    ) -> List[str]:
        """Launch new instances."""

        config: Dict[str, Any] = {}
        plans = self.get_plans()
        regions = self.list_regions()
        plans = [
            plan for plan in plans if plan['plan_id'] == instance_type and
            region in [r['id'] for r in plan['regions']]
        ]
        if not plans:
            raise FluidstackAPIError(
                f'Plan {instance_type} out of stock in region {region}')

        ssh_key = self.get_or_add_ssh_key(ssh_pub_key)
        os_id = 'Ubuntu 20.04 LTS'
        body = dict(plan=None if config else instance_type,
                    region=regions[region],
                    os=os_id,
                    hostname=hostname,
                    ssh_keys=[ssh_key['id']],
                    multiplicity=count,
                    config=config)

        response = requests.post(ENDPOINT + 'server',
                                 auth=(self.api_key, self.api_token),
                                 json=body)
        raise_fluidstack_error(response)
        instance_ids = response.json().get('multiple')
        assert all(id is not None for id in instance_ids), instance_ids
        return instance_ids

    def list_ssh_keys(self):
        response = requests.get(ENDPOINT + 'ssh',
                                auth=(self.api_key, self.api_token))
        raise_fluidstack_error(response)
        return response.json()

    def get_or_add_ssh_key(self, ssh_pub_key: str = '') -> Dict[str, str]:
        """Add ssh key if not already added."""
        ssh_keys = self.list_ssh_keys()
        for key in ssh_keys:
            if key['public_key'].strip() == ssh_pub_key.strip():
                return {
                    'id': key['id'],
                    'name': key['name'],
                    'ssh_key': ssh_pub_key
                }
        ssh_key_name = 'skypilot-' + get_key_suffix()
        response = requests.post(
            ENDPOINT + 'ssh',
            auth=(self.api_key, self.api_token),
            json=dict(name=ssh_key_name, public_key=ssh_pub_key),
        )
        raise_fluidstack_error(response)
        key_id = response.json()['id']
        return {'id': key_id, 'name': ssh_key_name, 'ssh_key': ssh_pub_key}

    @functools.lru_cache()
    def list_regions(self):
        response = requests.get(ENDPOINT + 'plans')
        raise_fluidstack_error(response)
        plans = response.json()
        plans = [
            plan for plan in plans
            if plan['minimum_commitment'] == 'hourly' and plan['type'] in
            ['preconfigured', 'custom'] and plan['gpu_type'] != 'NO GPU'
        ]

        def get_regions(plans: List) -> dict:
            """Return a list of regions where the plan is available."""
            regions = {}
            for plan in plans:
                for region in plan.get('regions', []):
                    regions[region['id']] = region['id']
            return regions

        regions = get_regions(plans)
        return regions

    def delete(self, instance_id: str):
        response = requests.delete(ENDPOINT + 'server/' + instance_id,
                                   auth=(self.api_key, self.api_token))
        raise_fluidstack_error(response)
        return response.json()

    def stop(self, instance_id: str):
        response = requests.put(ENDPOINT + 'server/' + instance_id + '/stop',
                                auth=(self.api_key, self.api_token))
        raise_fluidstack_error(response)
        return response.json()

    def restart(self, instance_id: str):
        response = requests.post(ENDPOINT + 'server/' + instance_id + '/reboot',
                                 auth=(self.api_key, self.api_token))
        raise_fluidstack_error(response)
        return response.json()

    def info(self, instance_id: str):
        response = requests.get(ENDPOINT + f'server/{instance_id}',
                                auth=(self.api_key, self.api_token))
        raise_fluidstack_error(response)
        return response.json()

    def status(self, instance_id: str):
        response = self.info(instance_id)
        return response['status']

    def add_tags(self, instance_id: str, tags: Dict[str, str]) -> str:
        response = requests.patch(
            ENDPOINT + f'server/{instance_id}/tag',
            auth=(self.api_key, self.api_token),
            json=dict(tags=json.dumps(tags)),
        )
        raise_fluidstack_error(response)
        return response.json()

    def rename(self, instance_id: str, hostname: str) -> str:
        response = requests.patch(
            ENDPOINT + f'server/{instance_id}/rename',
            auth=(self.api_key, self.api_token),
            json=dict(name=hostname),
        )
        raise_fluidstack_error(response)
        return response.json()



"""Cloud provision interface.

This module provides a standard low-level interface that all
providers supported by SkyPilot need to follow.
"""
import functools
import inspect
from typing import Any, Dict, List, Optional, Type

from sky import sky_logging
from sky import status_lib
# These provision.<cloud> modules should never fail even if underlying cloud SDK
# dependencies are not installed. This is ensured by using sky.adaptors inside
# these modules, for lazy loading of cloud SDKs.
from sky.provision import aws
from sky.provision import azure
from sky.provision import common
from sky.provision import cudo
from sky.provision import fluidstack
from sky.provision import gcp
from sky.provision import kubernetes
from sky.provision import runpod
from sky.provision import vsphere
from sky.utils import command_runner

logger = sky_logging.init_logger(__name__)


def _route_to_cloud_impl(func):

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        # check the signature to fail early
        inspect.signature(func).bind(*args, **kwargs)
        if args:
            provider_name = args[0]
            args = args[1:]
        else:
            provider_name = kwargs.pop('provider_name')

        module_name = provider_name.lower()
        module = globals().get(module_name)
        assert module is not None, f'Unknown provider: {module_name}'

        impl = getattr(module, func.__name__, None)
        if impl is not None:
            return impl(*args, **kwargs)

        # If implementation does not exist, fall back to default implementation
        return func(provider_name, *args, **kwargs)

    return _wrapper


# pylint: disable=unused-argument


@_route_to_cloud_impl
def query_instances(
    provider_name: str,
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    non_terminated_only: bool = True,
) -> Dict[str, Optional[status_lib.ClusterStatus]]:
    """Query instances.

    Returns a dictionary of instance IDs and status.

    A None status means the instance is marked as "terminated"
    or "terminating".
    """
    raise NotImplementedError


@_route_to_cloud_impl
def bootstrap_instances(
        provider_name: str, region: str, cluster_name_on_cloud: str,
        config: common.ProvisionConfig) -> common.ProvisionConfig:
    """Bootstrap configurations for a cluster.

    This function sets up auxiliary resources for a specified cluster
    with the provided configuration,
    and returns a ProvisionConfig object with updated configuration.
    These auxiliary resources could include security policies, network
    configurations etc. These resources tend to be free or very cheap,
    but it takes time to set them up from scratch. So we generally
    cache or reuse them when possible.
    """
    raise NotImplementedError


@_route_to_cloud_impl
def run_instances(provider_name: str, region: str, cluster_name_on_cloud: str,
                  config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Start instances with bootstrapped configuration."""
    raise NotImplementedError


@_route_to_cloud_impl
def stop_instances(
    provider_name: str,
    cluster_name_on_cloud: str,
    provider_config: Dict[str, Any],
    worker_only: bool = False,
) -> None:
    """Stop running instances."""
    raise NotImplementedError


@_route_to_cloud_impl
def terminate_instances(
    provider_name: str,
    cluster_name_on_cloud: str,
    provider_config: Dict[str, Any],
    worker_only: bool = False,
) -> None:
    """Terminate running or stopped instances."""
    raise NotImplementedError


@_route_to_cloud_impl
def open_ports(
    provider_name: str,
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    """Open ports for inbound traffic."""
    raise NotImplementedError


@_route_to_cloud_impl
def cleanup_ports(
    provider_name: str,
    cluster_name_on_cloud: str,
    # TODO: make ports optional and allow cleaning up only specified ports.
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    """Delete any opened ports."""
    raise NotImplementedError


@_route_to_cloud_impl
def query_ports(
    provider_name: str,
    cluster_name_on_cloud: str,
    ports: List[str],
    head_ip: Optional[str] = None,
    provider_config: Optional[Dict[str, Any]] = None,
) -> Dict[int, List[common.Endpoint]]:
    """Query details about ports on a cluster.

    If head_ip is provided, it may be used by the cloud implementation to
    return the endpoint without querying the cloud provider. If head_ip is not
    provided, the cloud provider will be queried to get the endpoint info.

    The underlying implementation is responsible for retries and timeout, e.g.
    kubernetes will wait for the service that expose the ports to be ready
    before returning the endpoint info.

    Returns a dict with port as the key and a list of common.Endpoint.
    """
    del provider_name, provider_config, cluster_name_on_cloud  # unused
    return common.query_ports_passthrough(ports, head_ip)


@_route_to_cloud_impl
def wait_instances(provider_name: str, region: str, cluster_name_on_cloud: str,
                   state: Optional[status_lib.ClusterStatus]) -> None:
    """Wait instances until they ends up in the given state."""
    raise NotImplementedError


@_route_to_cloud_impl
def get_cluster_info(
        provider_name: str,
        region: str,
        cluster_name_on_cloud: str,
        provider_config: Optional[Dict[str, Any]] = None) -> common.ClusterInfo:
    """Get the metadata of instances in a cluster."""
    raise NotImplementedError


@_route_to_cloud_impl
def get_command_runners(
    provider_name: str,
    cluster_info: common.ClusterInfo,
    **crednetials: Dict[str, Any],
) -> List[command_runner.CommandRunner]:
    """Get a command runner for the given cluster."""
    ip_list = cluster_info.get_feasible_ips()
    port_list = cluster_info.get_ssh_ports()
    return command_runner.SSHCommandRunner.make_runner_list(
        node_list=zip(ip_list, port_list),
        **crednetials,
    )



"""Benchmark utils: Utility functions to manage benchmarking process."""
import copy
import getpass
import glob
import json
from multiprocessing import pool
import os
import subprocess
import sys
import tempfile
import textwrap
import time
import typing
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import uuid

import colorama
import prettytable
from rich import progress as rich_progress

import sky
from sky import backends
from sky import data
from sky import global_user_state
from sky import sky_logging
from sky import status_lib
from sky.backends import backend_utils
from sky.benchmark import benchmark_state
from sky.data import storage as storage_lib
from sky.skylet import constants
from sky.skylet import job_lib
from sky.skylet import log_lib
from sky.utils import common_utils
from sky.utils import log_utils
from sky.utils import rich_utils
from sky.utils import subprocess_utils
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib

logger = sky_logging.init_logger(__name__)

_SKY_LOCAL_BENCHMARK_DIR = os.path.expanduser('~/.sky/benchmarks')
_SKY_REMOTE_BENCHMARK_DIR = '~/.sky/sky_benchmark_dir'
# NOTE: This must be the same as _SKY_REMOTE_BENCHMARK_DIR
# in sky/callbacks/sky_callback/base.py.
_SKY_REMOTE_BENCHMARK_DIR_SYMLINK = '~/sky_benchmark_dir'

# NOTE: This must be the same as _BENCHMARK_SUMMARY
# in sky/callbacks/sky_callback/base.py.
_BENCHMARK_SUMMARY = 'summary.json'
_RUN_START = 'run_start.txt'
_RUN_END = 'run_end.txt'

_Config = Dict[str, Any]


def _generate_cluster_names(benchmark: str, num_clusters: int) -> List[str]:
    if num_clusters == 1:
        names = [f'sky-bench-{benchmark}']
    else:
        names = [f'sky-bench-{benchmark}-{i}' for i in range(num_clusters)]
    for name in names:
        if global_user_state.get_cluster_from_name(name) is not None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Cluster name {name} is taken. '
                                 'Try using a different benchmark name.')
    return names


def _make_script_with_timelogs(script: str, start_path: str,
                               end_path: str) -> str:
    """Add prologue and epilogue that log the start and end times of the script.

    Using the logs, we can get the job status and duration even when the cluster
    has stopped or terminated. Note that the end time is only logged when the
    command finishes successfully.
    """
    return textwrap.dedent(f"""\
        echo $(date +%s.%N) > {start_path}
        {script}
        EXIT_CODE=$?
        if [ $EXIT_CODE -eq 0 ]; then
            echo $(date +%s.%N) > {end_path}
        fi
        exit $EXIT_CODE
        """)


def _get_optimized_resources(
        candidate_configs: List[_Config]) -> List['resources_lib.Resources']:
    candidate_configs = copy.deepcopy(candidate_configs)
    optimized_resources = []
    for config in candidate_configs:
        with sky.Dag() as dag:
            resources = config.get('resources', None)
            resources = sky.Resources.from_yaml_config(resources)
            task = sky.Task()
            task.set_resources(resources)

        dag = sky.optimize(dag, quiet=True)
        task = dag.tasks[0]
        optimized_resources.append(task.best_resources)
    return optimized_resources


def _print_candidate_resources(
        benchmark: str, clusters: List[str], config: _Config,
        candidate_resources: List['resources_lib.Resources']) -> None:
    task_str = config.get('name', 'a task')
    num_nodes = config.get('num_nodes', 1)
    logger.info(f'{colorama.Style.BRIGHT}Benchmarking {task_str} '
                f'on candidate resources (benchmark name: {benchmark}):'
                f'{colorama.Style.RESET_ALL}')

    columns = [
        'CLUSTER',
        'CLOUD',
        '# NODES',
        'INSTANCE',
        'vCPUs',
        'Mem(GB)',
        'ACCELERATORS',
        'PRICE ($/hr)',
    ]
    table_kwargs = {
        'hrules': prettytable.FRAME,
        'vrules': prettytable.NONE,
        'border': True,
    }
    candidate_table = log_utils.create_table(columns, **table_kwargs)

    for cluster, resources in zip(clusters, candidate_resources):
        if resources.accelerators is None:
            accelerators = '-'
        else:
            accelerator, count = list(resources.accelerators.items())[0]
            accelerators = f'{accelerator}:{count}'
        cloud = resources.cloud
        vcpus, mem = cloud.get_vcpus_mem_from_instance_type(
            resources.instance_type)

        def format_number(x):
            if x is None:
                return '-'
            elif x.is_integer():
                return str(int(x))
            else:
                return f'{x:.1f}'

        vcpus = format_number(vcpus)
        mem = format_number(mem)
        cost = num_nodes * resources.get_cost(3600)
        spot = '[Spot]' if resources.use_spot else ''
        row = [
            cluster, cloud, num_nodes, resources.instance_type + spot, vcpus,
            mem, accelerators, f'{cost:.2f}'
        ]
        candidate_table.add_row(row)
    logger.info(f'{candidate_table}\n')


def _create_benchmark_bucket() -> Tuple[str, str]:
    # Generate a bucket name.
    # TODO(woosuk): Use a more pleasant naming scheme.
    # TODO(woosuk): Ensure that the bucket name is globally unique.
    bucket_name = f'sky-bench-{uuid.uuid4().hex[:4]}-{getpass.getuser()}'

    # Select the bucket type.
    enabled_clouds = storage_lib.get_cached_enabled_storage_clouds_or_refresh(
        raise_if_no_cloud_access=True)
    # Already checked by raise_if_no_cloud_access=True.
    assert enabled_clouds
    bucket_type = data.StoreType.from_cloud(enabled_clouds[0]).value

    # Create a benchmark bucket.
    logger.info(f'Creating a bucket {bucket_name} to save the benchmark logs.')
    storage = data.Storage(bucket_name, source=None, persistent=True)
    storage.add_store(bucket_type)

    # Save the bucket name and type to the config.
    benchmark_state.set_benchmark_bucket(bucket_name, bucket_type)
    return bucket_name, bucket_type


def _format_err_msg(msg: str):
    return f'{colorama.Fore.RED}{msg}{colorama.Style.RESET_ALL}'


def _parallel_run_with_interrupt_handling(func: Callable,
                                          args: List[Any]) -> List[Any]:
    with pool.ThreadPool(processes=len(args)) as p:
        try:
            return list(p.imap(func, args))
        except KeyboardInterrupt:
            print()
            logger.error(_format_err_msg('Interrupted by user.'))
            subprocess_utils.run('sky status')
            sys.exit(1)


def _launch_with_log_suppress_exception(
        cluster: str, cmd: List[str],
        log_dir: str) -> Union[Tuple[int, str], Exception]:
    """Executes `sky launch` in a subprocess and returns normally.

    This function does not propagate any error so that failures in a
    launch thread do not disrupt the other parallel launch threads.
    """
    prefix_color = colorama.Fore.MAGENTA
    prefix = f'{prefix_color}({cluster}){colorama.Style.RESET_ALL} '
    try:
        returncode, _, stderr = log_lib.run_with_log(
            cmd,
            log_path=os.path.join(log_dir, f'{cluster}.log'),
            stream_logs=True,
            streaming_prefix=prefix,
            start_streaming_at='Creating a new cluster: ',
            skip_lines=[
                'Tip: to reuse an existing cluster, specify --cluster (-c).',
            ],
            end_streaming_at='Job submitted with Job ID: ',
            require_outputs=True,
        )
        # Report any error from the `sky launch` subprocess.
        return returncode, stderr
    except Exception as e:  # pylint: disable=broad-except
        # FIXME(woosuk): Avoid using general Exception.
        # Report any error in executing and processing the outputs of
        # the `sky launch` subprocess.
        return e


def _download_remote_dir(remote_dir: str, local_dir: str,
                         bucket_type: data.StoreType) -> None:
    # FIXME(woosuk): Replace this function with bucket.download_remote_dir.
    if bucket_type == data.StoreType.S3:
        remote_dir = f's3://{remote_dir}'
        subprocess.run(
            ['aws', 's3', 'cp', '--recursive', remote_dir, local_dir],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True)
    elif bucket_type == data.StoreType.GCS:
        remote_dir = f'gs://{remote_dir}'
        subprocess.run(['gsutil', '-m', 'cp', '-r', remote_dir, local_dir],
                       stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL,
                       check=True)
    else:
        raise RuntimeError('Azure Blob Storage is not supported yet.')


def _delete_remote_dir(remote_dir: str, bucket_type: data.StoreType) -> None:
    # FIXME(woosuk): Replace this function with bucket.delete_remote_dir.
    if bucket_type == data.StoreType.S3:
        remote_dir = f's3://{remote_dir}'
        subprocess.run(['aws', 's3', 'rm', '--recursive', remote_dir],
                       stdout=subprocess.DEVNULL,
                       stderr=subprocess.DEVNULL,
                       check=True)
    elif bucket_type == data.StoreType.GCS:
        remote_dir = f'gs://{remote_dir}'
        proc = subprocess.run(['gsutil', '-m', 'rm', '-r', remote_dir],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              check=False)
        if proc.returncode != 0:
            stderr = proc.stderr.decode('utf-8')
            if 'BucketNotFoundException: 404' in stderr:
                logger.warning(f'Bucket {remote_dir} does not exist. Skip')
            else:
                raise RuntimeError(f'Failed to delete {remote_dir}: {stderr}')
    else:
        raise RuntimeError('Azure Blob Storage is not supported yet.')


def _read_timestamp(path: str) -> float:
    with open(path, 'r', encoding='utf-8') as f:
        timestamp = f.readlines()
    assert len(timestamp) == 1
    return float(timestamp[0].strip())


def _update_benchmark_result(benchmark_result: Dict[str, Any]) -> Optional[str]:
    benchmark = benchmark_result['benchmark']
    benchmark_status = benchmark_result['status']
    cluster = benchmark_result['cluster']
    if benchmark_status.is_terminal():
        # No need to update.
        return

    # Get the start and end timestamps if exist.
    local_dir = os.path.join(_SKY_LOCAL_BENCHMARK_DIR, benchmark, cluster)
    run_start_path = os.path.join(local_dir, _RUN_START)
    start_time = None
    if os.path.exists(run_start_path):
        start_time = _read_timestamp(run_start_path)
    run_end_path = os.path.join(local_dir, _RUN_END)
    end_time = None
    if os.path.exists(run_end_path):
        # The job has terminated with a zero exit code. See
        # generate_benchmark_configs() which ensures the 'run' commands write
        # out end_time remotely on success; and the caller of this func which
        # downloads all benchmark log files including the end_time file to
        # local.
        end_time = _read_timestamp(run_end_path)

    # Get the status of the benchmarking cluster and job.
    record = global_user_state.get_cluster_from_name(cluster)
    cluster_status = None
    job_status = None
    if record is not None:
        cluster_status, handle = backend_utils.refresh_cluster_status_handle(
            cluster)
        if handle is not None:
            backend = backend_utils.get_backend_from_handle(handle)
            assert isinstance(backend, backends.CloudVmRayBackend)

            if cluster_status == status_lib.ClusterStatus.UP:
                # NOTE: The id of the benchmarking job must be 1.
                # TODO(woosuk): Handle exceptions.
                job_status = backend.get_job_status(handle,
                                                    job_ids=[1],
                                                    stream_logs=False)[1]

    logger.debug(f'Cluster {cluster}, cluster_status: {cluster_status}, '
                 f'benchmark_status {benchmark_status}, job_status: '
                 f'{job_status}, start_time {start_time}, end_time {end_time}')

    # Update the benchmark status.
    if end_time is not None:
        # The job has terminated with zero exit code.
        benchmark_status = benchmark_state.BenchmarkStatus.FINISHED
    elif cluster_status is None:
        # Candidate cluster: preempted or never successfully launched.
        #
        # Note that benchmark record is only inserted after all clusters
        # finished launch() (successful or not). See
        # launch_benchmark_clusters(). So this case doesn't include "just before
        # candidate cluster's launch() is called".

        # See above: if cluster_status is not UP, job_status is defined as None.
        assert job_status is None, job_status
        benchmark_status = benchmark_state.BenchmarkStatus.TERMINATED
    elif cluster_status == status_lib.ClusterStatus.INIT:
        # Candidate cluster's launch has something gone wrong, or is still
        # launching.

        # See above: if cluster_status is not UP, job_status is defined as None.
        assert job_status is None, job_status
        benchmark_status = benchmark_state.BenchmarkStatus.INIT
    elif cluster_status == status_lib.ClusterStatus.STOPPED:
        # Candidate cluster is auto-stopped, or user manually stops it at any
        # time. Also, end_time is None.

        # See above: if cluster_status is not UP, job_status is defined as None.
        assert job_status is None, job_status
        benchmark_status = benchmark_state.BenchmarkStatus.TERMINATED
    else:
        assert cluster_status == status_lib.ClusterStatus.UP, (
            'ClusterStatus enum should have been handled')
        if job_status is None:
            benchmark_status = benchmark_state.BenchmarkStatus.INIT
        else:
            if job_status < job_lib.JobStatus.RUNNING:
                benchmark_status = benchmark_state.BenchmarkStatus.INIT
            elif job_status == job_lib.JobStatus.RUNNING:
                benchmark_status = benchmark_state.BenchmarkStatus.RUNNING
            else:
                assert job_status.is_terminal(), '> RUNNING means terminal'
                # Case: cluster_status UP, job_status.is_terminal()
                if job_status == job_lib.JobStatus.SUCCEEDED:
                    # Since we download the benchmark logs before checking the
                    # cluster status, there is a chance that the end timestamp
                    # is saved and the cluster is stopped AFTER we download the
                    # logs.  In this case, we consider the current timestamp as
                    # the end time.
                    end_time = time.time()
                    benchmark_status = benchmark_state.BenchmarkStatus.FINISHED
                else:
                    benchmark_status = (
                        benchmark_state.BenchmarkStatus.TERMINATED)

    callback_log_dirs = glob.glob(os.path.join(local_dir, 'sky-callback-*'))
    if callback_log_dirs:
        # There can be multiple logs if the cluster has executed multiple jobs.
        # Here, we consider the first log as the log of the benchmarking job.
        log_dir = sorted(callback_log_dirs)[0]
        summary_path = os.path.join(log_dir, _BENCHMARK_SUMMARY)
    else:
        summary_path = None

    message = None
    if summary_path is not None and os.path.exists(summary_path):
        # (1) SkyCallback has saved the summary.
        with open(summary_path, 'r', encoding='utf-8') as f:
            summary = json.load(f)
        if end_time is None:
            last_time = summary['last_step_time']
        else:
            last_time = end_time
        if last_time is None:
            if job_status == job_lib.JobStatus.RUNNING:
                last_time = time.time()
            else:
                message = (f'No duration information found for {cluster}. '
                           'Check if at least 1 step has finished.')
        record = benchmark_state.BenchmarkRecord(
            start_time=start_time,
            last_time=last_time,
            num_steps_so_far=summary['num_steps'],
            seconds_per_step=summary['time_per_step'],
            estimated_total_seconds=summary['estimated_total_time'],
        )
    elif end_time is not None:
        # (2) The benchmarking job has terminated normally
        # without SkyCallback logs.
        record = benchmark_state.BenchmarkRecord(start_time=start_time,
                                                 last_time=end_time)
    elif job_status == job_lib.JobStatus.RUNNING:
        # (3) SkyCallback is not initialized yet or not used.
        message = ('SkyCallback is not initialized yet '
                   f'or not used for {cluster}.')
        record = benchmark_state.BenchmarkRecord(start_time=start_time,
                                                 last_time=time.time())
    elif benchmark_status == benchmark_state.BenchmarkStatus.TERMINATED:
        # (4) The benchmarking job has terminated abnormally.
        message = (f'The benchmarking job on {cluster} has terminated with '
                   'non-zero exit code.')
        record = benchmark_state.BenchmarkRecord(start_time=start_time,
                                                 last_time=None)
    else:
        # (5) Otherwise (e.g., cluster_status is INIT).
        message = f'No benchmark logs found for {cluster}.'
        record = benchmark_state.BenchmarkRecord(start_time=None,
                                                 last_time=None)
    benchmark_state.update_benchmark_result(benchmark, cluster,
                                            benchmark_status, record)
    return message


def generate_benchmark_configs(
    benchmark: str,
    config: _Config,
    candidates: List[Dict[str, str]],
) -> Tuple[List[str], List[_Config]]:
    # Generate a config for each cluster.
    clusters = _generate_cluster_names(benchmark, len(candidates))
    candidate_configs = []
    # TODO(woosuk): Use a jinja template.
    for cluster, candidate in zip(clusters, candidates):
        # Re-override the config with each candidate config.
        candidate_config = copy.deepcopy(config)
        if 'resources' not in candidate_config:
            candidate_config['resources'] = {}
        if 'candidates' in candidate_config['resources']:
            del candidate_config['resources']['candidates']
        candidate_config['resources'].update(candidate)

        # Mount the benchmark bucket to SKY_BENCHMARK_DIR.
        if 'file_mounts' not in candidate_config:
            candidate_config['file_mounts'] = {}
        # The bucket name and type are specified at launch time.
        candidate_config['file_mounts'][_SKY_REMOTE_BENCHMARK_DIR] = {
            'name': None,
            'mode': 'MOUNT',
            'store': None,
        }

        benchmark_dir = os.path.join(_SKY_REMOTE_BENCHMARK_DIR, benchmark,
                                     cluster)
        if 'setup' not in candidate_config:
            candidate_config['setup'] = ''
        # Create a symbolic link to a directory in the benchmark bucket.
        candidate_config['setup'] = textwrap.dedent(f"""\
            mkdir -p {benchmark_dir}
            ln -s {benchmark_dir} {_SKY_REMOTE_BENCHMARK_DIR_SYMLINK}
            {candidate_config['setup']}""")

        # Log the start and end time of the benchmarking task.
        if 'run' not in candidate_config:
            candidate_config['run'] = ''
        candidate_config['run'] = _make_script_with_timelogs(
            candidate_config['run'], os.path.join(benchmark_dir, _RUN_START),
            os.path.join(benchmark_dir, _RUN_END))

        candidate_configs.append(candidate_config)
    return clusters, candidate_configs


def print_benchmark_clusters(benchmark: str, clusters: List[str],
                             config: _Config,
                             candidate_configs: List[_Config]) -> None:
    candidate_resources = _get_optimized_resources(candidate_configs)
    _print_candidate_resources(benchmark, clusters, config, candidate_resources)


def launch_benchmark_clusters(benchmark: str, clusters: List[str],
                              candidate_configs: List[_Config],
                              commandline_args: List[Dict[str, Any]]) -> bool:
    # Use a Sky storage to save the benchmark logs.
    bucket_name, bucket_type = benchmark_state.get_benchmark_bucket()
    if bucket_name is not None:
        handle = global_user_state.get_handle_from_storage_name(bucket_name)
        if handle is not None:
            assert bucket_type is not None

    # If the bucket does not exist, create one.
    if bucket_name is None or handle is None:
        bucket_name, bucket_type = _create_benchmark_bucket()

    # Remove the previous benchmark logs if exist.
    remove_benchmark_logs(benchmark, bucket_name, data.StoreType[bucket_type])

    # The benchmark bucket is mounted to _SKY_REMOTE_BENCHMARK_DIR.
    for candidate_config in candidate_configs:
        bucket_config = candidate_config['file_mounts'][
            _SKY_REMOTE_BENCHMARK_DIR]
        bucket_config['name'] = bucket_name
        bucket_config['store'] = bucket_type

    # Generate a temporary yaml file for each cluster.
    yaml_fds = []
    for cluster, candidate_config in zip(clusters, candidate_configs):
        # pylint: disable=consider-using-with
        f = tempfile.NamedTemporaryFile('w',
                                        prefix=f'{cluster}-',
                                        suffix='.yaml')
        common_utils.dump_yaml(f.name, candidate_config)
        yaml_fds.append(f)
        logger.debug(f'Generated temporary yaml file: {f.name}')

    # Generate a common launch command.
    cmd = ['-d', '-y']
    for arg_name, arg in commandline_args.items():
        if isinstance(arg, list):
            # 'env' arguments.
            for v in arg:
                cmd += [f'--{arg_name}', str(v)]
        else:
            cmd += [f'--{arg_name}', str(arg)]

    # Generate a launch command for each cluster.
    launch_cmds = [['sky', 'launch', yaml_fd.name, '-c', cluster] + cmd
                   for yaml_fd, cluster in zip(yaml_fds, clusters)]

    # Save stdout/stderr from cluster launches.
    run_timestamp = backend_utils.get_run_timestamp()
    log_dir = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp)
    log_dir = os.path.expanduser(log_dir)
    logger.info(
        f'{colorama.Fore.YELLOW}To view stdout/stderr from individual '
        f'cluster launches, check the logs in{colorama.Style.RESET_ALL} '
        f'{colorama.Style.BRIGHT}{log_dir}/{colorama.Style.RESET_ALL}')

    # Launch the benchmarking clusters in parallel.
    outputs = _parallel_run_with_interrupt_handling(
        lambda args: _launch_with_log_suppress_exception(*args, log_dir=log_dir
                                                        ),
        list(zip(clusters, launch_cmds)))

    # Handle the errors raised during the cluster launch.
    for cluster, output in zip(clusters, outputs):
        if isinstance(output, Exception):
            logger.error(_format_err_msg(f'Launching {cluster} failed.'))
            logger.error(output)
        else:
            returncode, stderr = output
            if returncode != 0:
                message = _format_err_msg(
                    f'Launching {cluster} failed with code {returncode}.')
                logger.error(message)
                logger.error(stderr)

    # Delete the temporary yaml files.
    for f in yaml_fds:
        f.close()

    # If at least one cluster has been provisioned (in whatever state),
    # add the benchmark to the state so that `sky bench down` can
    # terminate the launched clusters.
    benchmark_created = False
    for cluster in clusters:
        record = global_user_state.get_cluster_from_name(cluster)
        if record is not None:
            if not benchmark_created:
                task_name = candidate_configs[0].get('name', None)
                benchmark_state.add_benchmark(benchmark, task_name, bucket_name)
                benchmark_created = True
            benchmark_state.add_benchmark_result(benchmark, record['handle'])
    return benchmark_created


def update_benchmark_state(benchmark: str) -> None:
    benchmark_results = benchmark_state.get_benchmark_results(benchmark)
    if all(result['status'].is_terminal() for result in benchmark_results):
        return

    bucket_name = benchmark_state.get_benchmark_from_name(benchmark)['bucket']
    handle = global_user_state.get_handle_from_storage_name(bucket_name)
    bucket_type = list(handle.sky_stores.keys())[0]

    # Download the benchmark logs from the benchmark bucket.
    # FIXME(woosuk): Do not download the logs if not necessary.
    remote_dir = os.path.join(bucket_name, benchmark)
    local_dir = os.path.join(_SKY_LOCAL_BENCHMARK_DIR, benchmark)
    os.makedirs(local_dir, exist_ok=True)
    with rich_utils.safe_status('[bold cyan]Downloading benchmark logs[/]'):
        _download_remote_dir(remote_dir, local_dir, bucket_type)

    # Update the benchmark results in parallel.
    num_candidates = len(benchmark_results)
    plural = 's' if num_candidates > 1 else ''
    progress = rich_progress.Progress(transient=True,
                                      redirect_stdout=False,
                                      redirect_stderr=False)
    task = progress.add_task(
        f'[bold cyan]Processing {num_candidates} benchmark result{plural}[/]',
        total=num_candidates)

    def _update_with_progress_bar(arg: Any) -> None:
        message = _update_benchmark_result(arg)
        if message is None:
            progress.update(task, advance=1)
        else:
            progress.stop()
            logger.info(
                f'{colorama.Fore.YELLOW}{message}{colorama.Style.RESET_ALL}')
            progress.start()

    with progress:
        _parallel_run_with_interrupt_handling(_update_with_progress_bar,
                                              benchmark_results)
        progress.live.transient = False
        progress.refresh()


def remove_benchmark_logs(benchmark: str, bucket_name: str,
                          bucket_type: data.StoreType) -> None:
    # Delete logs in the benchmark bucket.
    remote_dir = os.path.join(bucket_name, benchmark)
    _delete_remote_dir(remote_dir, bucket_type)
    # Delete logs in the local storage.
    local_dir = os.path.join(_SKY_LOCAL_BENCHMARK_DIR, benchmark)
    subprocess.run(['rm', '-rf', local_dir], check=False)
